@@ -1,32 +1,21 @@
 /* 
  * math.h
+ * This file has no copyright assigned and is placed in the Public Domain.
+ * This file is a part of the mingw-runtime package.
+ * No warranty is given; refer to the file DISCLAIMER within the package.
  *
  * Mathematical functions.
  *
- * This file is part of the Mingw32 package.
- *
- * Contributors:
- *  Created by Colin Peters <colin@bird.fu.is.saga-u.ac.jp>
- *
- *  THIS SOFTWARE IS NOT COPYRIGHTED
- *
- *  This source code is offered for use in the public domain. You may
- *  use, modify or distribute it freely.
- *
- *  This code is distributed in the hope that it will be useful but
- *  WITHOUT ANY WARRANTY. ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
- *  DISCLAIMED. This includes but is not limited to warranties of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
- *
- * $Revision: 1.16 $
- * $Author: earnie $
- * $Date: 2003/10/10 15:02:04 $
- *
  */
 
+
 #ifndef _MATH_H_
 #define _MATH_H_
 
+#if __GNUC__ >= 3
+#pragma GCC system_header
+#endif
+
 /* All the headers include this file. */
 #include <_mingw.h>
 
@@ -145,7 +134,6 @@ struct _exception
 	double	retval;
 };
 
-
 _CRTIMP double __cdecl sin (double);
 _CRTIMP double __cdecl cos (double);
 _CRTIMP double __cdecl tan (double);
@@ -159,10 +147,7 @@ _CRTIMP double __cdecl atan2 (double, double);
 _CRTIMP double __cdecl exp (double);
 _CRTIMP double __cdecl log (double);
 _CRTIMP double __cdecl log10 (double);
-#ifdef __NO_ISOCEXT
-_CRTIMP
-#endif
-	double __cdecl pow (double, double);
+_CRTIMP	double __cdecl pow (double, double);
 _CRTIMP double __cdecl sqrt (double);
 _CRTIMP double __cdecl ceil (double);
 _CRTIMP double __cdecl floor (double);
@@ -172,10 +157,50 @@ _CRTIMP double __cdecl frexp (double, int*);
 _CRTIMP double __cdecl modf (double, double*);
 _CRTIMP double __cdecl fmod (double, double);
 
+/* Excess precision when using a 64-bit mantissa for FPU math ops can
+   cause unexpected results with some of the MSVCRT math functions.  For
+   example, unless the function return value is stored (truncating to
+   53-bit mantissa), calls to pow with both x and y as integral values
+   sometimes produce a non-integral result.
+   One workaround is to reset the FPU env to 53-bit mantissa
+   by a call to fesetenv (FE_PC53_ENV).  Amother is to force storage
+   of the return value of individual math functions using wrappers.
+   NB, using these wrappers will disable builtin math functions and
+   hence disable the folding of function results at compile time when
+   arguments are constant.  */
+
+#if 0
+#define __DEFINE_FLOAT_STORE_MATHFN_D1(fn1)	\
+static __inline__ double			\
+__float_store_ ## fn1 (double x)		\
+{						\
+   __volatile__ double res = (fn1) (x);		\
+  return res;					\
+}
+
+#define __DEFINE_FLOAT_STORE_MATHFN_D2(fn2)	\
+static __inline__ double			\
+__float_store_ ## fn2 (double x, double y)	\
+{						\
+  __volatile__ double res = (fn2) (x, y);	\
+  return res;					\
+}
+#endif
+
+/* For example, here is how to force the result of the pow function
+   to be stored:   */
+#if 0
+#undef pow
+/* Define the ___float_store_pow function and use it instead of pow().  */
+__DEFINE_FLOAT_STORE_MATHFN_D2 (pow)
+#define pow __float_store_pow
+#endif
 
 #ifndef __STRICT_ANSI__
 
-/* Complex number (for cabs). This really belongs in complex.h */
+/* Complex number (for _cabs). This is the MS version. The ISO
+   C99 counterpart _Complex is an intrinsic type in GCC and
+   'complex' is defined as a macro.  See complex.h  */
 struct _complex
 {
 	double	x;	/* Real part */
@@ -220,7 +245,6 @@ _CRTIMP int __cdecl _isnan (double);
 
 #if !defined (_NO_OLDNAMES)
 
-_CRTIMP double __cdecl cabs (struct _complex);
 _CRTIMP double __cdecl j0 (double);
 _CRTIMP double __cdecl j1 (double);
 _CRTIMP double __cdecl jn (int, double);
@@ -233,14 +257,31 @@ _CRTIMP double __cdecl scalb (double, long);
 _CRTIMP int __cdecl finite (double);
 _CRTIMP int __cdecl fpclass (double);
 
+#define FP_SNAN    _FPCLASS_SNAN
+#define FP_QNAN    _FPCLASS_QNAN
+#define FP_NINF    _FPCLASS_NINF
+#define FP_PINF    _FPCLASS_PINF
+#define FP_NDENORM _FPCLASS_ND
+#define FP_PDENORM _FPCLASS_PD
+#define FP_NZERO   _FPCLASS_NZ
+#define FP_PZERO   _FPCLASS_PZ
+#define FP_NNORM   _FPCLASS_NN
+#define FP_PNORM   _FPCLASS_PN
+
 #endif /* Not _NO_OLDNAMES */
 
+/* This require msvcr70.dll or higher. */ 
+#if __MSVCRT_VERSION__ >= 0x0700
+_CRTIMP int __cdecl _set_SSE2_enable (int);
+#endif /* __MSVCRT_VERSION__ >= 0x0700 */
+
+
 #endif /* __STRICT_ANSI__ */
 
 
 #ifndef __NO_ISOCEXT
 #if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L) \
-	|| !defined __STRICT_ANSI__ || defined __GLIBCPP__
+	|| !defined __STRICT_ANSI__ || defined __cplusplus
 
 #define NAN (0.0F/0.0F)
 #define HUGE_VALF (1.0F/0.0F)
@@ -332,19 +373,19 @@ __CRT_INLINE int __cdecl __isnanl (long double _x)
 __CRT_INLINE int __cdecl __signbit (double x) {
   unsigned short stw;
   __asm__ ( "fxam; fstsw %%ax;": "=a" (stw) : "t" (x));
-  return stw & 0x0200;
+  return (stw & 0x0200) != 0;
 }
 
 __CRT_INLINE int __cdecl __signbitf (float x) {
   unsigned short stw;
   __asm__ ("fxam; fstsw %%ax;": "=a" (stw) : "t" (x));
-  return stw & 0x0200;
+  return (stw & 0x0200) != 0;
 }
 
 __CRT_INLINE int __cdecl __signbitl (long double x) {
   unsigned short stw;
   __asm__ ("fxam; fstsw %%ax;": "=a" (stw) : "t" (x));
-  return stw & 0x0200;
+  return (stw & 0x0200) != 0;
 }
 
 #define signbit(x) (sizeof (x) == sizeof (float) ? __signbitf (x)	\
@@ -386,10 +427,23 @@ __CRT_INLINE float __cdecl tanhf (float x)
   {return (float) tanh (x);}
 extern long double __cdecl tanhl (long double);
 
-/*
- * TODO: asinh, acosh, atanh
- */ 
+/* Inverse hyperbolic trig functions  */ 
+/* 7.12.5.1 */
+extern double __cdecl acosh (double);
+extern float __cdecl acoshf (float);
+extern long double __cdecl acoshl (long double);
 
+/* 7.12.5.2 */
+extern double __cdecl asinh (double);
+extern float __cdecl asinhf (float);
+extern long double __cdecl asinhl (long double);
+
+/* 7.12.5.3 */
+extern double __cdecl atanh (double);
+extern float __cdecl atanf  (float);
+extern long double __cdecl atanhl (long double);
+
+/* Exponentials and logarithms  */
 /* 7.12.6.1 Double in C89 */
 __CRT_INLINE float __cdecl expf (float x)
   {return (float) exp (x);}
@@ -400,7 +454,11 @@ extern double __cdecl exp2(double);
 extern float __cdecl exp2f(float);
 extern long double __cdecl exp2l(long double);
 
-/* 7.12.6.3 The expm1 functions: TODO */
+/* 7.12.6.3 The expm1 functions */
+/* TODO: These could be inlined */
+extern double __cdecl expm1(double);
+extern float __cdecl expm1f(float);
+extern long double __cdecl expm1l(long double);
 
 /* 7.12.6.4 Double in C89 */
 __CRT_INLINE float __cdecl frexpf (float x, int* expn)
@@ -442,6 +500,9 @@ extern double __cdecl logb (double);
 extern float __cdecl logbf (float);
 extern long double __cdecl logbl (long double);
 
+/* Inline versions.  GCC-4.0+ can do a better fast-math optimization
+   with __builtins. */ 
+#if !(__MINGW_GNUC_PREREQ (4, 0) && defined __FAST_MATH__ )
 __CRT_INLINE double __cdecl logb (double x)
 {
   double res;
@@ -465,6 +526,7 @@ __CRT_INLINE long double __cdecl logbl (long double x)
        "fstp	%%st" : "=t" (res) : "0" (x));
   return res;
 }
+#endif /* !defined __FAST_MATH__ || !__MINGW_GNUC_PREREQ (4, 0) */
 
 /* 7.12.6.12  Double in C89 */
 extern float __cdecl modff (float, float*);
@@ -507,16 +569,12 @@ extern long double __cdecl sqrtl (long double);
 /* 7.12.8.1 The erf functions  */
 extern double __cdecl erf (double);
 extern float __cdecl erff (float);
-/* TODO
 extern long double __cdecl erfl (long double);
-*/ 
 
 /* 7.12.8.2 The erfc functions  */
 extern double __cdecl erfc (double);
 extern float __cdecl erfcf (float);
-/* TODO
 extern long double __cdecl erfcl (long double);
-*/ 
 
 /* 7.12.8.3 The lgamma functions */
 extern double __cdecl lgamma (double);
@@ -543,6 +601,22 @@ extern long double __cdecl nearbyintl (long double);
 
 /* 7.12.9.4 */
 /* round, using fpu control word settings */
+extern double __cdecl rint (double);
+extern float __cdecl rintf (float);
+extern long double __cdecl rintl (long double);
+
+/* 7.12.9.5 */
+extern long __cdecl lrint (double);
+extern long __cdecl lrintf (float);
+extern long __cdecl lrintl (long double);
+
+extern long long __cdecl llrint (double);
+extern long long __cdecl llrintf (float);
+extern long long __cdecl llrintl (long double);
+
+/* Inline versions of above. 
+   GCC 4.0+ can do a better fast-math job with __builtins. */
+#if !(__MINGW_GNUC_PREREQ (4, 0) && defined __FAST_MATH__ )
 __CRT_INLINE double __cdecl rint (double x)
 {
   double retval;
@@ -564,54 +638,54 @@ __CRT_INLINE long double __cdecl rintl (long double x)
   return retval;
 }
 
-/* 7.12.9.5 */
 __CRT_INLINE long __cdecl lrint (double x) 
 {
   long retval;  
-  __asm__ __volatile__							      \
-    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
 
 __CRT_INLINE long __cdecl lrintf (float x) 
 {
   long retval;
-  __asm__ __volatile__							      \
-    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
 
 __CRT_INLINE long __cdecl lrintl (long double x) 
 {
   long retval;
-  __asm__ __volatile__							      \
-    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpl %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
 
-__CRT_INLINE long long __cdecl llrint (double x) 
+__CRT_INLINE long long __cdecl llrint (double x)
 {
   long long retval;
-  __asm__ __volatile__							      \
-    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
 
-__CRT_INLINE long long __cdecl llrintf (float x) 
+__CRT_INLINE long long __cdecl llrintf (float x)
 {
   long long retval;
-  __asm__ __volatile__							      \
-    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
 
 __CRT_INLINE long long __cdecl llrintl (long double x) 
 {
   long long retval;
-  __asm__ __volatile__							      \
-    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");				      \
+  __asm__ __volatile__
+    ("fistpll %0"  : "=m" (retval) : "t" (x) : "st");
   return retval;
 }
+#endif /* !__FAST_MATH__ || !__MINGW_GNUC_PREREQ (4,0)  */
 
 /* 7.12.9.6 */
 /* round away from zero, regardless of fpu control word settings */
@@ -667,10 +741,12 @@ extern long double __cdecl nanl(const char *tagp);
 /* 7.12.11.3 */
 extern double __cdecl nextafter (double, double); /* in libmoldname.a */
 extern float __cdecl nextafterf (float, float);
-/* TODO: Not yet implemented */
-/* extern long double __cdecl nextafterl (long double, long double); */
+extern long double __cdecl nextafterl (long double, long double);
 
-/* 7.12.11.4 The nexttoward functions: TODO */
+/* 7.12.11.4 The nexttoward functions */
+extern double __cdecl nexttoward (double,  long double);
+extern float __cdecl nexttowardf (float,  long double);
+extern long double __cdecl nexttowardl (long double, long double);
 
 /* 7.12.12.1 */
 /*  x > y ? (x - y) : 0.0  */
